@@ -1,5 +1,5 @@
 /*
- *   Copyright (C) 2009-2011 Erwin Waterlander
+ *   Copyright (C) 2009-2012 Erwin Waterlander
  *   All rights reserved.
  * 
  *   Redistribution and use in source and binary forms, with or without
@@ -75,13 +75,24 @@
 #include <locale.h>
 #endif
 #include <errno.h>
+#ifdef D2U_UNICODE
+#include <wchar.h>
+#endif
 
 #if (defined(__WATCOMC__) && defined(__NT__))  /* Watcom */
-#  define WIN32
+#  define WIN32 1
+#endif
+
+#if defined(__WATCOMC__) && defined(__I86__) /* Watcom C, 16 bit Intel */
+#define MSDOS 1
+#endif
+
+#if defined(__WATCOMC__) && defined(__DOS__) /* Watcom C, 32 bit DOS */
+#define MSDOS 1
 #endif
 
 #if defined(WIN32) && !defined(__CYGWIN__) /* Windows */
-#define MSDOS
+#define MSDOS 1
 #endif
 
 #if  defined(__TURBOC__) || defined(DJGPP) || defined(__MINGW32__)
@@ -133,6 +144,9 @@
 #define OUTPUTFILE_SYMLINK 0x8
 #define INPUT_TARGET_NO_REGFILE 0x10
 #define OUTPUT_TARGET_NO_REGFILE 0x20
+#define LOCALE_NOT_UTF8 0x40
+#define WCHAR_T_TOO_SMALL 0x80
+#define UNICODE_CONVERSION_ERROR 0x100
 
 #define CONVMODE_ASCII  0
 #define CONVMODE_7BIT   1
@@ -152,6 +166,11 @@
 #define SYMLINK_FOLLOW 1
 #define SYMLINK_REPLACE 2
 
+#define FILE_MBS     0  /* Multi-byte string or 8-bit char */
+#define FILE_UTF16LE 1  /* UTF-16 Little Endian */
+#define FILE_UTF16BE 2  /* UTF-16 Big Endian */
+#define FILE_UTF8    3  /* UTF-8 */
+
 typedef struct
 {
   int NewFile;                          /* is in new file mode? */
@@ -165,6 +184,8 @@ typedef struct
   int status;
   int stdio_mode;                       /* if TRUE, stdio mode */
   int error;                            /* an error occurred */
+  int bomtype;                          /* byte order mark */
+  int add_bom;                          /* 1: write BOM */
 } CFlag;
 
 
@@ -188,4 +209,10 @@ FILE* MakeTempFileFrom(const char *OutFN, char **fname_ret);
 int MakeTempFileFrom(const char *OutFN, char **fname_ret);
 #endif
 int ResolveSymbolicLink(char *lFN, char **rFN, CFlag *ipFlag, char *progname);
+FILE *read_bom (FILE *f, int *bomtype);
+#ifdef D2U_UNICODE
+wint_t d2u_getwc(FILE *f, int bomtype);
+wint_t d2u_ungetwc(wint_t wc, FILE *f, int bomtype);
+wint_t d2u_putwc(wint_t wc, FILE *f, CFlag *ipFlag);
+#endif
 
