@@ -60,14 +60,16 @@
  */
 
 
-#define VER_AUTHOR   "Erwin Waterlander"
-
 /* #define DEBUG 1 */
 
 #include "common.h"
 #include "dos2unix.h"
 #include "querycp.h"
-
+#ifdef D2U_UNICODE
+#ifndef MSDOS  /* Unix, Cygwin */
+# include <langinfo.h>
+#endif
+#endif
 
 void PrintLicense(void)
 {
@@ -80,6 +82,27 @@ All rights reserved.\n\n"));
   PrintBSDLicense();
 }
 
+#ifdef D2U_UNICODE
+void StripDelimiterW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, wint_t CurChar)
+{
+  wint_t TempNextChar;
+  /* CurChar is always CR (x0d) */
+  /* In normal dos2unix mode put nothing (skip CR). */
+  /* Don't modify Mac files when in dos2unix mode. */
+  if ( (TempNextChar = d2u_getwc(ipInF, ipFlag->bomtype)) != WEOF) {
+    d2u_ungetwc( TempNextChar, ipInF, ipFlag->bomtype);  /* put back peek char */
+    if ( TempNextChar != 0x0a ) {
+      d2u_putwc( CurChar, ipOutF, ipFlag);  /* Mac line, put back CR */
+    }
+  }
+  else if ( CurChar == 0x0d ) {  /* EOF: last Mac line delimiter (CR)? */
+    d2u_putwc( CurChar, ipOutF, ipFlag);
+  }
+  if (ipFlag->NewLine) {  /* add additional LF? */
+    d2u_putwc(0x0a, ipOutF, ipFlag);
+  }
+}
+#endif
 
 void StripDelimiter(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, int CurChar)
 {
@@ -87,20 +110,140 @@ void StripDelimiter(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, int CurChar)
   /* CurChar is always CR (x0d) */
   /* In normal dos2unix mode put nothing (skip CR). */
   /* Don't modify Mac files when in dos2unix mode. */
-  if ( (TempNextChar = getc(ipInF)) != EOF) {
+  if ( (TempNextChar = fgetc(ipInF)) != EOF) {
     ungetc( TempNextChar, ipInF );  /* put back peek char */
     if ( TempNextChar != '\x0a' ) {
-      putc( CurChar, ipOutF );  /* Mac line, put back CR */
+      fputc( CurChar, ipOutF );  /* Mac line, put back CR */
     }
   }
   else if ( CurChar == '\x0d' ) {  /* EOF: last Mac line delimiter (CR)? */
-    putc( CurChar, ipOutF );
+    fputc( CurChar, ipOutF );
   }
   if (ipFlag->NewLine) {  /* add additional LF? */
-    putc('\x0a', ipOutF);
+    fputc('\x0a', ipOutF);
   }
 }
 
+/* converts stream ipInF to UNIX format text and write to stream ipOutF
+ * RetVal: 0  if success
+ *         -1  otherwise
+ */
+#ifdef D2U_UNICODE
+int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, char *progname)
+{
+    int RetVal = 0;
+    wint_t TempChar;
+    wint_t TempNextChar;
+
+    ipFlag->status = 0;
+
+    /* CR-LF -> LF */
+    /* LF    -> LF, in case the input file is a Unix text file */
+    /* CR    -> CR, in dos2unix mode (don't modify Mac file) */
+    /* CR    -> LF, in Mac mode */
+    /* \x0a = Newline/Line Feed (LF) */
+    /* \x0d = Carriage Return (CR) */
+
+    switch (ipFlag->FromToMode)
+    {
+      case FROMTO_DOS2UNIX: /* dos2unix */
+        while ((TempChar = d2u_getwc(ipInF, ipFlag->bomtype)) != WEOF) {  /* get character */
+          if ((ipFlag->Force == 0) &&
+              (TempChar < 32) &&
+              (TempChar != 0x0a) &&  /* Not an LF */
+              (TempChar != 0x0d) &&  /* Not a CR */
+              (TempChar != 0x09) &&  /* Not a TAB */
+              (TempChar != 0x0c)) {  /* Not a form feed */
+            RetVal = -1;
+            ipFlag->status |= BINARY_FILE ;
+            break;
+          }
+          if (TempChar != 0x0d) {
+            if (d2u_putwc(TempChar, ipOutF, ipFlag) == WEOF) {
+              RetVal = -1;
+              if (!ipFlag->Quiet)
+              {
+                if (!(ipFlag->status & UNICODE_CONVERSION_ERROR))
+                {
+                  fprintf(stderr, "%s: ", progname);
+                  fprintf(stderr, "%s", _("can not write to output file\n"));
+                }
+              }
+              break;
+            } 
+          } else {
+            StripDelimiterW( ipInF, ipOutF, ipFlag, TempChar );
+          }
+        }
+        break;
+      case FROMTO_MAC2UNIX: /* mac2unix */
+        while ((TempChar = d2u_getwc(ipInF, ipFlag->bomtype)) != WEOF) {
+          if ((ipFlag->Force == 0) &&
+              (TempChar < 32) &&
+              (TempChar != 0x0a) &&  /* Not an LF */
+              (TempChar != 0x0d) &&  /* Not a CR */
+              (TempChar != 0x09) &&  /* Not a TAB */
+              (TempChar != 0x0c)) {  /* Not a form feed */
+            RetVal = -1;
+            ipFlag->status |= BINARY_FILE ;
+            break;
+          }
+          if ((TempChar != 0x0d))
+            {
+              if(d2u_putwc(TempChar, ipOutF, ipFlag) == WEOF){
+                RetVal = -1;
+                if (!ipFlag->Quiet)
+                {
+                  if (!(ipFlag->status & UNICODE_CONVERSION_ERROR))
+                  {
+                    fprintf(stderr, "%s: ", progname);
+                    fprintf(stderr, "%s", _("can not write to output file\n"));
+                  }
+                }
+                break;
+              }
+            }
+          else{
+            /* TempChar is a CR */
+            if ( (TempNextChar = d2u_getwc(ipInF, ipFlag->bomtype)) != WEOF) {
+              ungetc( TempNextChar, ipInF );  /* put back peek char */
+              /* Don't touch this delimiter if it's a CR,LF pair. */
+              if ( TempNextChar == 0x0a ) {
+                d2u_putwc(0x0d, ipOutF, ipFlag); /* put CR, part of DOS CR-LF */
+                continue;
+              }
+            }
+            if (d2u_putwc(0x0a, ipOutF, ipFlag) == WEOF) /* MAC line end (CR). Put LF */
+              {
+                RetVal = -1;
+                if (!ipFlag->Quiet)
+                {
+                  if (!(ipFlag->status & UNICODE_CONVERSION_ERROR))
+                  {
+                    fprintf(stderr, "%s: ", progname);
+                    fprintf(stderr, "%s", _("can not write to output file\n"));
+                  }
+                }
+                break;
+              }
+            if (ipFlag->NewLine) {  /* add additional LF? */
+              d2u_putwc(0x0a, ipOutF, ipFlag);
+            }
+          }
+        }
+        break;
+      default: /* unknown FromToMode */
+      ;
+#if DEBUG
+      fprintf(stderr, "%s: ", progname);
+      fprintf(stderr, _("program error, invalid conversion mode %d\n"),ipFlag->FromToMode);
+      exit(1);
+#endif
+    }
+    return RetVal;
+}
+#endif
+
 /* converts stream ipInF to UNIX format text and write to stream ipOutF
  * RetVal: 0  if success
  *         -1  otherwise
@@ -160,7 +303,7 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, char *progname)
     switch (ipFlag->FromToMode)
     {
       case FROMTO_DOS2UNIX: /* dos2unix */
-        while ((TempChar = getc(ipInF)) != EOF) {  /* get character */
+        while ((TempChar = fgetc(ipInF)) != EOF) {  /* get character */
           if ((ipFlag->Force == 0) &&
               (TempChar < 32) &&
               (TempChar != '\x0a') &&  /* Not an LF */
@@ -172,7 +315,7 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, char *progname)
             break;
           }
           if (TempChar != '\x0d') {
-            if (putc(ConvTable[TempChar], ipOutF) == EOF) {
+            if (fputc(ConvTable[TempChar], ipOutF) == EOF) {
               RetVal = -1;
               if (!ipFlag->Quiet)
               {
@@ -187,7 +330,7 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, char *progname)
         }
         break;
       case FROMTO_MAC2UNIX: /* mac2unix */
-        while ((TempChar = getc(ipInF)) != EOF) {
+        while ((TempChar = fgetc(ipInF)) != EOF) {
           if ((ipFlag->Force == 0) &&
               (TempChar < 32) &&
               (TempChar != '\x0a') &&  /* Not an LF */
@@ -200,7 +343,7 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, char *progname)
           }
           if ((TempChar != '\x0d'))
             {
-              if(putc(ConvTable[TempChar], ipOutF) == EOF){
+              if(fputc(ConvTable[TempChar], ipOutF) == EOF){
                 RetVal = -1;
                 if (!ipFlag->Quiet)
                 {
@@ -212,15 +355,15 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, char *progname)
             }
           else{
             /* TempChar is a CR */
-            if ( (TempNextChar = getc(ipInF)) != EOF) {
+            if ( (TempNextChar = fgetc(ipInF)) != EOF) {
               ungetc( TempNextChar, ipInF );  /* put back peek char */
               /* Don't touch this delimiter if it's a CR,LF pair. */
               if ( TempNextChar == '\x0a' ) {
-                putc('\x0d', ipOutF); /* put CR, part of DOS CR-LF */
+                fputc('\x0d', ipOutF); /* put CR, part of DOS CR-LF */
                 continue;
               }
             }
-            if (putc('\x0a', ipOutF) == EOF) /* MAC line end (CR). Put LF */
+            if (fputc('\x0a', ipOutF) == EOF) /* MAC line end (CR). Put LF */
               {
                 RetVal = -1;
                 if (!ipFlag->Quiet)
@@ -231,7 +374,7 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, char *progname)
                 break;
               }
             if (ipFlag->NewLine) {  /* add additional LF? */
-              putc('\x0a', ipOutF);
+              fputc('\x0a', ipOutF);
             }
           }
         }
@@ -370,9 +513,62 @@ int ConvertDosToUnixNewFile(char *ipInFN, char *ipOutFN, CFlag *ipFlag, char *pr
     }
   }
 
+  InF = read_bom(InF, &ipFlag->bomtype);
+
+#ifdef D2U_UNICODE
+#ifndef MSDOS  /* Unix, Cygwin */
+  if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE))
+  {
+    if (strcmp(nl_langinfo(CODESET), "UTF-8") != 0)
+    {
+      /* Don't convert UTF-16 files when the locale encoding is not UTF-8
+       * to prevent loss of characters. */
+      ipFlag->status |= LOCALE_NOT_UTF8 ;
+      if (!ipFlag->error) ipFlag->error = 1;
+      RetVal = -1;
+    }
+  }
+#endif
+#if !defined(WIN32) && !defined(__CYGWIN__) /* Not Windows or Cygwin */
+  if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE))
+  {
+    if (sizeof(wchar_t) < 4)
+    {
+      /* A decoded UTF-16 surrogate pair must fit in a wchar_t */
+      ipFlag->status |= WCHAR_T_TOO_SMALL ;
+      if (!ipFlag->error) ipFlag->error = 1;
+      RetVal = -1;
+    }
+  }
+#endif
+#endif
+
+  if (ipFlag->add_bom)
+    fprintf(TempF, "%s", "\xEF\xBB\xBF");  /* UTF-8 BOM */
+
+  /* Turn off ISO and 7-bit conversion for Unicode text files */
+  if (ipFlag->bomtype > 0)
+    ipFlag->ConvMode = CONVMODE_ASCII;
+
   /* conversion sucessful? */
+#ifdef D2U_UNICODE
+  if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE))
+  {
+    if ((!RetVal) && (ConvertDosToUnixW(InF, TempF, ipFlag, progname)))
+      RetVal = -1;
+    if (ipFlag->status & UNICODE_CONVERSION_ERROR)
+    {
+      if (!ipFlag->error) ipFlag->error = 1;
+      RetVal = -1;
+    }
+  } else {
+    if ((!RetVal) && (ConvertDosToUnix(InF, TempF, ipFlag, progname)))
+      RetVal = -1;
+  }
+#else
   if ((!RetVal) && (ConvertDosToUnix(InF, TempF, ipFlag, progname)))
     RetVal = -1;
+#endif
 
    /* can close in file? */
   if ((InF) && (fclose(InF) == EOF))
@@ -553,11 +749,23 @@ int ConvertDosToUnixStdio(CFlag *ipFlag, char *progname)
 
     _setmode(fileno(stdout), O_BINARY);
     _setmode(fileno(stdin), O_BINARY);
-    return (ConvertDosToUnix(stdin, stdout, ipFlag, progname));
 #elif defined(MSDOS) || defined(__CYGWIN__) || defined(__OS2__)
     setmode(fileno(stdout), O_BINARY);
     setmode(fileno(stdin), O_BINARY);
-    return (ConvertDosToUnix(stdin, stdout, ipFlag, progname));
+#endif
+
+    read_bom(stdin, &ipFlag->bomtype);
+
+    if (ipFlag->add_bom)
+       fprintf(stdout, "%s", "\xEF\xBB\xBF");  /* UTF-8 BOM */
+
+#ifdef D2U_UNICODE
+    if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE))
+    {
+       return (ConvertDosToUnixW(stdin, stdout, ipFlag, progname));
+    } else {
+       return (ConvertDosToUnix(stdin, stdout, ipFlag, progname));
+    }
 #else
     return (ConvertDosToUnix(stdin, stdout, ipFlag, progname));
 #endif
@@ -578,6 +786,9 @@ int main (int argc, char *argv[])
 #ifdef ENABLE_NLS
   char localedir[1024];
 #endif
+# ifdef __MINGW64__
+  int _dowildcard = -1; /* enable wildcard expansion for Win64 */
+# endif
 
   progname[8] = '\0';
   strcpy(progname,"dos2unix");
@@ -620,6 +831,10 @@ int main (int argc, char *argv[])
   pFlag->status = 0;
   pFlag->stdio_mode = 1;
   pFlag->error = 0;
+#ifdef D2U_UNICODE
+  pFlag->bomtype = FILE_MBS;
+#endif
+  pFlag->add_bom = 0;
 
   if ( ((ptr=strrchr(argv[0],'/')) == NULL) && ((ptr=strrchr(argv[0],'\\')) == NULL) )
     ptr = argv[0];
@@ -655,6 +870,8 @@ int main (int argc, char *argv[])
         pFlag->Quiet = 1;
       else if ((strcmp(argv[ArgIdx],"-l") == 0) || (strcmp(argv[ArgIdx],"--newline") == 0))
         pFlag->NewLine = 1;
+      else if ((strcmp(argv[ArgIdx],"-m") == 0) || (strcmp(argv[ArgIdx],"--add-bom") == 0))
+        pFlag->add_bom = 1;
       else if ((strcmp(argv[ArgIdx],"-S") == 0) || (strcmp(argv[ArgIdx],"--skip-symlink") == 0))
         pFlag->Follow = SYMLINK_SKIP;
       else if ((strcmp(argv[ArgIdx],"-F") == 0) || (strcmp(argv[ArgIdx],"--follow-symlink") == 0))
@@ -829,6 +1046,27 @@ int main (int argc, char *argv[])
               fprintf(stderr,"%s: ",progname);
               fprintf(stderr, _("code page %d is not supported.\n"), pFlag->ConvMode);
             }
+          } else if (pFlag->status & LOCALE_NOT_UTF8)
+          {
+            if (!pFlag->Quiet)
+            {
+              fprintf(stderr,"%s: ",progname);
+              fprintf(stderr, _("Skipping UTF-16 file %s, the current locale character encoding is not UTF-8.\n"), argv[ArgIdx-1]);
+            }
+          } else if (pFlag->status & WCHAR_T_TOO_SMALL)
+          {
+            if (!pFlag->Quiet)
+            {
+              fprintf(stderr,"%s: ",progname);
+              fprintf(stderr, _("Skipping UTF-16 file %s, the size of wchar_t is %d bytes.\n"), argv[ArgIdx-1], (int)sizeof(wchar_t));
+            }
+          } else if (pFlag->status & UNICODE_CONVERSION_ERROR)
+          {
+            if (!pFlag->Quiet)
+            {
+              fprintf(stderr,"%s: ",progname);
+              fprintf(stderr, _("Skipping UTF-16 file %s, an UTF-16 conversion error occurred.\n"), argv[ArgIdx-1]);
+            }
           } else {
             if (!pFlag->Quiet)
             {
@@ -885,6 +1123,27 @@ int main (int argc, char *argv[])
             fprintf(stderr,"%s: ",progname);
             fprintf(stderr, _("code page %d is not supported.\n"), pFlag->ConvMode);
           }
+        } else if (pFlag->status & LOCALE_NOT_UTF8)
+        {
+          if (!pFlag->Quiet)
+          {
+            fprintf(stderr,"%s: ",progname);
+            fprintf(stderr, _("Skipping UTF-16 file %s, the current locale character encoding is not UTF-8.\n"), argv[ArgIdx]);
+          }
+        } else if (pFlag->status & WCHAR_T_TOO_SMALL)
+        {
+          if (!pFlag->Quiet)
+          {
+            fprintf(stderr,"%s: ",progname);
+            fprintf(stderr, _("Skipping UTF-16 file %s, the size of wchar_t is %d bytes.\n"), argv[ArgIdx], (int)sizeof(wchar_t));
+          }
+        } else if (pFlag->status & UNICODE_CONVERSION_ERROR)
+        {
+          if (!pFlag->Quiet)
+          {
+            fprintf(stderr,"%s: ",progname);
+            fprintf(stderr, _("Skipping UTF-16 file %s, an UTF-16 conversion error occurred.\n"), argv[ArgIdx]);
+          }
         } else {
           if (!pFlag->Quiet)
           {
